from ctypes import CDLL, POINTER, CFUNCTYPE, c_int, c_long, c_longlong, c_double, c_void_p, c_char, c_char_p, pythonapi, py_object, sizeof
from ctypes import util

import numpy as np
import os

if sizeof(c_long) == 8:
    c_int64 = c_long
else:
    c_int64 = c_longlong

DPTR = POINTER(c_double)
CPTRPTR = POINTER(c_char_p)

FFUNC = CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, c_void_p)
JFUNC = CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, c_void_p)
JVFUNC= CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, c_int64, DPTR, c_void_p)
OFUNC = CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, DPTR,
                  CPTRPTR, CPTRPTR, c_int64, c_void_p)

def find_uopt():
    path = util.find_library("uopt")
    if path is None:
        if os.name == 'nt':
            path = 'uopt.dll'
        else:
            path = 'libuopt.so'
    return path

def fix_opts(opts):
    if not 'enableExit' in opts:
        opts['enableExit'] = 0
    if not 'enableSignalHandler' in opts:
        opts['enableSignalHandler'] = 0
    if not 'enableFloatExcept' in opts:
        opts['enableFloatExcept'] = 0
    return opts


def uopt(x0, xs, ys, f, g, H=None, Hv=None, Jv=None, s=None, opts={}, udata=None):
    return uopt0(x0, xs, ys, f, g, H, Hv, Jv, s, opts, udata)

def uopt0(x0, xs, ys, fhandle, ghandle, Hhandle, Hvhandle, Jvhandle, ohandle, opts, udata):

    mkarr = np.ctypeslib.as_array

    def fstub(x, nx, y, ny, userdata):
        fhandle(mkarr(x,(nx,)), mkarr(y,(ny,)), udata)
        return 0

    def gstub(x, n, y, ny, g, userdata):
        ghandle(mkarr(x,(n,)), mkarr(y,(ny,)), mkarr(g,(n,)), udata)
        return 0

    def Jacvstub(x, nx, y, ny, dx, ndx, Jv, userdata):
        if ndx > 1:
            raise("not implemented")
        Jvhandle(mkarr(x,(nx,)), mkarr(y,(ny,)), mkarr(dx,(nx,ndx)), mkarr(Jv,(ny,ndx)), udata)
        return 0

    def Hstub(x, n, y, ny, g, H, userdata):
        Hhandle(mkarr(x,(n,)), mkarr(y,(ny,)), mkarr(g,(n,)), mkarr(H,(n,n)), udata)
        return 0

    def Hvstub(x, n, y, ny, dx, ndx, g, H, userdata):
        Hvhandle(mkarr(x,(n,)), mkarr(y,(ny,)), mkarr(dx,(n,ndx)), mkarr(g,(n,)), mkarr(H,(n,ndx)), udata)
        return 0

    def ostub(x, nx, y, ny, g, H, onames, ovals, no, userdata):
        if ohandle is None:
            return 0
        d = {onames[i].decode('utf8'): ovals[i].decode('utf8') for i in range(no)}
        r = ohandle(mkarr(x,(nx,)), mkarr(y,(ny,)), d, udata)
        return int(r) if r is not None else 0

    path = find_uopt()
    try:
        dll = CDLL(path)
    except BaseException as E:
        print("Failed to find libuopt", E)
        raise(E)
    cuopt = dll.uopt

    opts = fix_opts(opts)

    names  = [bytes(k, 'utf-8') for k in opts.keys()]
    values = [float(k) for k in opts.values()]

    HFUNC = CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, DPTR, c_void_p)
    HVFUNC= CFUNCTYPE(c_int, DPTR, c_int64, DPTR, c_int64, DPTR, c_int64, DPTR, DPTR, c_void_p)

    cuopt.return_type = c_int
    cuopt.argtypes = (DPTR, DPTR, c_int64, DPTR, c_int64,
                      FFUNC, JFUNC, JVFUNC, HFUNC, HVFUNC, OFUNC,
                      CPTRPTR, DPTR, c_int64, c_void_p)
    cuopt.errcheck = my_errcheck

    ffunc  = FFUNC(fstub)
    gfunc  = JFUNC(gstub)
    Jvfunc = JVFUNC(Jacvstub if Jvhandle is not None else 0)
    if Hhandle is None:
        Hfunc = HFUNC(0)
    else:
        Hfunc  = HFUNC(Hstub)
    if Hvhandle is None:
        Hvfunc = HVFUNC(0)
    else:
        Hvfunc = HVFUNC(Hvstub)
    ofunc  = OFUNC(ostub)

    cx0 = np.ctypeslib.as_ctypes(x0)
    cxs = np.ctypeslib.as_ctypes(xs)
    cys = np.ctypeslib.as_ctypes(ys)
    STRARR = c_char_p * len(names)
    DBLARR = c_double * len(names)
    cnames = STRARR()
    cvalues = DBLARR()
    cnames[:] = names
    cvalues[:] = values
    res = cuopt(cx0, cxs, x0.size, cys, ys.size,
                ffunc, gfunc, Jvfunc, Hfunc, Hvfunc, ofunc,
                cnames, cvalues, len(names), None)
    return res

def usolve(x0, xs, ys, f, J, Jv=None, JTv=None, s=None, opts={}, udata=None):
    return usolve0(x0, xs, ys, f, J, Jv, JTv, s, opts, udata)

def usolve0(x0, xs, ys, fhandle, Jhandle, Jvhandle, JTvhandle, ohandle, opts, udata):

    mkarr = np.ctypeslib.as_array
    runinfo = {'status': {}}

    def fstub(x, nx, y, ny, userdata):
        fhandle(mkarr(x, (nx,)), mkarr(y, (ny,)), udata)
        return 0

    def Jacstub(x, nx, y, ny, Jac, userdata):
        Jarr = mkarr(Jac, (ny, nx))
        Jarr = np.lib.stride_tricks.as_strided(Jarr, strides=(sizeof(c_double), ny*sizeof(c_double)))
        Jhandle(mkarr(x, (nx,)), mkarr(y, (ny,)), Jarr, udata)
        return 0

    def Jacvstub(x, nx, y, ny, dx, ndx, Jv, userdata):
        if ndx > 1:
            raise("not implemented")
        Jvhandle(mkarr(x, (nx,)), mkarr(y, (ny,)), mkarr(dx, (nx,ndx)), mkarr(Jv, (ny,ndx)), udata)
        return 0

    def JacTvstub(x, nx, y, ny, dy, ndy, JTv, userdata):
        if ndy > 1:
            raise("not implemented")
        JTvhandle(mkarr(x, (nx,)), mkarr(y, (ny,)), mkarr(dy, (ndy, ny)), mkarr(JTv, (ndy, nx)), udata)
        return 0

    def ostub(x, nx, y, ny, J, _, onames, ovals, no, userdata):
        if ohandle is None:
            return 0
        d = {onames[i].decode('utf8'): ovals[i].decode('utf8') for i in range(no)}
        r = ohandle(mkarr(x, (nx,)), mkarr(y, (ny,)), d, udata)
        if 'exit' in d:
            runinfo['status'] = d
        return int(r) if r is not None else 0

    path = find_uopt()
    try:
        dll = CDLL(path)
    except BaseException as E:
        print("Failed to find libuopt", E)
        raise(E)

    cusolve = dll.usolve

    opts = fix_opts(opts)

    names  = [bytes(k, 'utf-8') for k in opts.keys()]
    values = [float(k) for k in opts.values()]

    cusolve.return_type = c_int
    cusolve.argtypes = (DPTR, DPTR, c_int64, DPTR, c_int64,
                        FFUNC, JFUNC, JVFUNC, JVFUNC, c_void_p, OFUNC,
                        CPTRPTR, DPTR, c_int64, c_void_p)
    cusolve.errcheck = my_errcheck

    ffunc  = FFUNC(fstub)
    Jfunc  = JFUNC(Jacstub if Jhandle is not None else 0)
    if Jvhandle is None:
        Jvfunc = JVFUNC(0)
    else:
        Jvfunc = JVFUNC(Jacvstub)
    JTvfunc  = JVFUNC(JacTvstub if JTvhandle is not None else 0)
    ofunc  = OFUNC(ostub)

    cx0 = np.ctypeslib.as_ctypes(x0)
    cxs = np.ctypeslib.as_ctypes(xs)
    cys = np.ctypeslib.as_ctypes(ys)
    STRARR = c_char_p * len(names)
    DBLARR = c_double * len(names)
    cnames = STRARR()
    cvalues = DBLARR()
    cnames[:] = names
    cvalues[:] = values

    res = cusolve(cx0, cxs, len(x0), cys, len(ys),
                  ffunc, Jfunc, Jvfunc, JTvfunc, None, ofunc,
                  cnames, cvalues, len(cnames), c_void_p())

    return (res, runinfo)

def my_errcheck(result, func, args):
    #print('my_errcheck %s: %d' %(func.__name__, result))
    return result
