#include <stdlib.h>
#include <R_ext/Rdynload.h>
#include <R_ext/Visibility.h>  // optional
#include <R.h>
#include <Rinternals.h>
#include <iostream>
#include <sstream>
#include <vector>

#include <uopt.h>

SEXP mkans(double x)
{
    SEXP ans = allocVector(REALSXP, 1);
    REAL(ans)[0] = x;
    return ans;
}

SEXP wrapuopt(SEXP x0,
              SEXP fHandle, SEXP gHandle, SEXP JvHandle, SEXP HHandle, SEXP HvHandle,
              SEXP oHandle, SEXP options, SEXP env)
{
  if (!isVector(x0)) {
    error("x0 must be a vector");
  }
  if (!isReal(x0)) {
    error("x0 must be a vector of reals");
  }
  if (!isFunction(fHandle)) {
    error("f must be a function");
  }
  if (!isFunction(gHandle)) {
    error("g must be a function");
  }
  if (!(isNull(JvHandle) || isFunction(JvHandle))) {
    error("Jv must be a function or null");
  }
  if (!(isNull(HHandle) || isFunction(HHandle))) {
    error("H must be a function or null");
  }
  if (!(isNull(HvHandle) || isFunction(HvHandle))) {
    std::cout << "Hv: " << isNull(HvHandle) << ", " << isFunction(HvHandle) << "\n";
    error("Hv must be a function or null");
  }
  if (!isFunction(oHandle)) {
    error("s must be a function");
  }
  if (!isNewList(options)) {
    error("options must be a list");
  }
  if (!isEnvironment(env)) {
    error("env must be an environment");
  }

  R_len_t const n = length(x0);
  R_len_t const m = 1;

  SEXP onames = getAttrib(options, R_NamesSymbol);
  SEXP elt = R_NilValue;

  std::vector<char const*> myoptnames;
  std::vector<double> myoptvals;

  for (R_len_t i = 0; i < length(options); ++i) {
    char const *name = CHAR(STRING_ELT(onames, i));
    elt = VECTOR_ELT(options, i);
    if (!isReal(elt)) {
      std::ostringstream errmsg;
      errmsg << "Options value " << i << " (" << name << ") must be real scalar\n";
      error(errmsg.str().c_str());
    }
    elt = PROTECT(coerceVector(elt, REALSXP));
    if (*name != '\0') {
      myoptnames.push_back(name);
      myoptvals.push_back(REAL(elt)[0]);
    }
  }

  SEXP R_fcall = PROTECT(lang2(fHandle, R_NilValue));
  auto fhClosure = [&R_fcall, &env](double const *x, R_len_t n, double *J) -> int {
    SEXP xk = PROTECT(allocVector(REALSXP, n));
    memcpy(REAL(xk), x, sizeof(double)*n);

    SETCADR(R_fcall, xk);
    SEXP ans = eval(R_fcall, env);

    // if (!isReal(ans)) {
    //   error("Function f must return a scalar real");
    // }
    ans = PROTECT(coerceVector(ans, REALSXP));

    *J = REAL(ans)[0];
    UNPROTECT(2);

    return 0;
  };


  SEXP R_gcall = PROTECT(lang2(gHandle, R_NilValue));
  auto ghClosure = [&R_gcall, &env](double const *x, R_len_t n, double *J, double *g) -> int {
    SEXP xk = PROTECT(allocVector(REALSXP, n));
    memcpy(REAL(xk), x, sizeof(double)*n);

    SETCADR(R_gcall, xk);

    SEXP ans = eval(R_gcall, env);

    if (!isNewList(ans)) {
      error("Function g must return a list");
    }
    if (length(ans) < 2) {
      error("Function g must return a list of length 2");
    }

    SEXP elt1 = VECTOR_ELT(ans, 0);
    SEXP elt2 = VECTOR_ELT(ans, 1);

    if (length(elt1) != n) {
      error("First element of function g's return value must be a real vector of length n");
    }
    if (!isReal(elt2)) {
      error("Second element of function g's return value must be a real scalar");
    }

    elt1 = PROTECT(coerceVector(elt1, REALSXP));
    elt2 = PROTECT(coerceVector(elt2, REALSXP));

    memcpy(g, REAL(elt1), sizeof(double)*n);
    *J = REAL(elt2)[0];

    UNPROTECT(3);

    return 0;
  };


  SEXP R_Jvcall = PROTECT(lang3(JvHandle, R_NilValue, R_NilValue));
  auto JvhClosure = [&R_Jvcall, &env](double const *x, R_len_t n, double *J,
                                     double const *dx, R_len_t ndx, double *Jv) -> int {
    SEXP xk = PROTECT(allocVector(REALSXP, n));
    memcpy(REAL(xk), x, sizeof(double)*n);

    SEXP dxk = PROTECT(allocMatrix(REALSXP, n, ndx));
    memcpy(REAL(dxk), dx, sizeof(double)*n*ndx);

    SEXP nextArg = CDR(R_Jvcall);
    SETCAR(nextArg, xk);
    nextArg = CDR(nextArg);
    SETCAR(nextArg, dxk);

    SEXP ans = eval(R_Jvcall, env);

    if (!isNewList(ans)) {
      error("Function Jv must return a list");
    }
    if (length(ans) < 3) {
      error("Function Jv must return a list of length 2");
    }

    SEXP elt1 = VECTOR_ELT(ans, 0);
    SEXP elt2 = VECTOR_ELT(ans, 1);

    if (length(elt1) != n) {
      error("First element of function Jv's return value must be a real vector of length n");
    }
    if (!isReal(elt2)) {
      error("Third element of function H's return value must be a real scalar");
    }

    elt1 = PROTECT(coerceVector(elt1, REALSXP));
    elt2 = PROTECT(coerceVector(elt2, REALSXP));

    *Jv = REAL(elt1)[0];
    *J = REAL(elt2)[0];

    UNPROTECT(4);

    return 0;
  };


  SEXP R_Hcall = PROTECT(lang2(HHandle, R_NilValue));
  auto HhClosure = [&R_Hcall, &env](double const *x, R_len_t n, double *J, double *g, double *H) -> int {
    SEXP xk = PROTECT(allocVector(REALSXP, n));
    memcpy(REAL(xk), x, sizeof(double)*n);

    SETCADR(R_Hcall, xk);

    SEXP ans = eval(R_Hcall, env);

    if (!isNewList(ans)) {
      error("Function H must return a list");
    }
    if (length(ans) < 3) {
      error("Function H must return a list of length 3");
    }

    SEXP elt1 = VECTOR_ELT(ans, 0);
    SEXP elt2 = VECTOR_ELT(ans, 1);
    SEXP elt3 = VECTOR_ELT(ans, 2);

    if (length(elt1) != n*n) {
      error("First element of function H's return value must be a real vector of length n^2");
    }
    if (length(elt2) != n) {
      error("Second element of function H's return value must be a real vector of length n");
    }
    if (!isReal(elt3)) {
      error("Third element of function H's return value must be a real scalar");
    }

    elt1 = PROTECT(coerceVector(elt1, REALSXP));
    elt2 = PROTECT(coerceVector(elt2, REALSXP));
    elt3 = PROTECT(coerceVector(elt3, REALSXP));

    memcpy(H, REAL(elt1), sizeof(double)*n*n);
    memcpy(g, REAL(elt2), sizeof(double)*n);
    *J = REAL(elt3)[0];

    UNPROTECT(4);

    return 0;
  };


  SEXP R_Hvcall = PROTECT(lang3(HvHandle, R_NilValue, R_NilValue));
  auto HvhClosure = [&R_Hvcall, &env](double const *x, R_len_t n, double const *dx, R_len_t ndx, double *J, double *g, double *H) -> int {
    SEXP xk = PROTECT(allocVector(REALSXP, n));
    memcpy(REAL(xk), x, sizeof(double)*n);

    SEXP dxk = PROTECT(allocMatrix(REALSXP, n, ndx));
    memcpy(REAL(dxk), dx, sizeof(double)*n*ndx);

    SEXP nextArg = CDR(R_Hvcall);
    SETCAR(nextArg, xk);
    nextArg = CDR(nextArg);
    SETCAR(nextArg, dxk);

    SEXP ans = eval(R_Hvcall, env);

    if (!isNewList(ans)) {
      error("Function H must return a list");
    }
    if (length(ans) < 3) {
      error("Function H must return a list of length 3");
    }

    SEXP elt1 = VECTOR_ELT(ans, 0);
    SEXP elt2 = VECTOR_ELT(ans, 1);
    SEXP elt3 = VECTOR_ELT(ans, 2);

    if (length(elt1) != n*ndx) {
      error("First element of function Hv's return value must be a real vector of length n*ndx");
    }
    if (length(elt2) != n) {
      error("Second element of function Hv's return value must be a real vector of length n");
    }
    if (!isReal(elt3)) {
      error("Third element of function Hv's return value must be a real scalar");
    }

    elt1 = PROTECT(coerceVector(elt1, REALSXP));
    elt2 = PROTECT(coerceVector(elt2, REALSXP));
    elt3 = PROTECT(coerceVector(elt3, REALSXP));

    memcpy(H, REAL(elt1), sizeof(double)*n*ndx);
    memcpy(g, REAL(elt2), sizeof(double)*n);
    *J = REAL(elt3)[0];

    UNPROTECT(5);

    return 0;
  };

  static SEXP lastStatus0, lastStatus1;

  SEXP R_ocall = PROTECT(lang4(oHandle, R_NilValue, R_NilValue, R_NilValue));
  auto ohClosure = [&R_ocall, &env](double const *x, R_len_t n,
                                    double const *y, R_len_t m,
                                    char const **onames, char const **ovals,
                                    int64_t no) -> int {
    SEXP xk = PROTECT(allocVector(REALSXP, n));
    memcpy(REAL(xk), x, sizeof(double)*n);
    SEXP yk = PROTECT(allocVector(REALSXP, m));
    memcpy(REAL(yk), y, sizeof(double)*m);

    SEXP sk     = PROTECT(allocVector(VECSXP, no));
    SEXP snames = PROTECT(allocVector(VECSXP, no));

    for (int64_t i = 0; i < no; ++i) {
      SEXP sval = PROTECT(allocVector(STRSXP, 1));
      SET_STRING_ELT(sval, 0, mkChar(ovals[i]));
      SET_VECTOR_ELT(sk, i, sval);

      sval = PROTECT(allocVector(STRSXP, 1));
      SET_STRING_ELT(sval, 0, mkChar(onames[i]));
      SET_VECTOR_ELT(snames, i, sval);

    }
    UNPROTECT(2 * no);

    setAttrib(sk, R_NamesSymbol, snames);

    lastStatus1 = lastStatus0;
    lastStatus0 = sk;

    SEXP nextArg = CDR(R_ocall);
    SETCAR(nextArg, xk);
    nextArg = CDR(nextArg);
    SETCAR(nextArg, yk);
    nextArg = CDR(nextArg);
    SETCAR(nextArg, sk);

    SEXP ans = eval(R_ocall, env);

    int oreturncode = 0;

    if (isNull(ans)) {
      UNPROTECT(4);
    } else {
      if (!isReal(ans)) {
        error("Function sh must return a scalar real");
      }
      ans = PROTECT(coerceVector(ans, REALSXP));
      oreturncode = REAL(ans)[0];
      UNPROTECT(5);
    }

    return oreturncode;
  };


  struct CallInfo {
    void *f1, *g1, *Jv1, *h1, *hv1, *o1;
  } mycallInfo;
  mycallInfo.f1 = &fhClosure;
  mycallInfo.g1 = &ghClosure;
  mycallInfo.Jv1 = &JvhClosure;
  mycallInfo.h1 = &HhClosure;
  mycallInfo.hv1 = &HvhClosure;
  mycallInfo.o1 = &ohClosure;

  fhandle fh = [](double const *x, int64_t n, double *y, int64_t, void *arg) -> int {
        CallInfo *ci = static_cast<CallInfo*>(arg);
        auto fhClosLoc = static_cast<decltype(fhClosure)*>(ci->f1);
        (*fhClosLoc)(x, n, y);
        return 0;
  };

  Jhandle gh = [](double const *x, int64_t n, double *y, int64_t, double *g, void *arg) -> int {
        CallInfo *ci = static_cast<CallInfo*>(arg);
        auto ghClosLoc = static_cast<decltype(ghClosure)*>(ci->g1);
        (*ghClosLoc)(x, n, y, g);
        return 0;
  };

  Jvhandle Jvh = 0;
  if (!isNull(JvHandle)) {
    Jvh = [](double const *x, int64_t n, double *y, int64_t, double const *dx, int64_t ndx, double *Jv, void *arg) -> int {
           CallInfo *ci = static_cast<CallInfo*>(arg);
           auto JvhClosLoc = static_cast<decltype(JvhClosure)*>(ci->Jv1);
           (*JvhClosLoc)(x, n, y, dx, ndx, Jv);
           return 0;
         };
  }

  Hhandle Hh = 0;
  if (!isNull(HHandle)) {
    Hh = [](double const *x, int64_t n, double *y, int64_t, double *g, double *H, void *arg) -> int {
           CallInfo *ci = static_cast<CallInfo*>(arg);
           auto HhClosLoc = static_cast<decltype(HhClosure)*>(ci->h1);
           (*HhClosLoc)(x, n, y, g, H);
           return 0;
         };
  }

  Hvhandle Hvh = 0;
  if (!isNull(HvHandle)) {
    Hvh = [](double const *x, int64_t n, double *y, int64_t,
             double const *dx, int64_t ndx, double *g, double *H, void *arg) -> int {
            CallInfo *ci = static_cast<CallInfo*>(arg);
            auto HvhClosLoc = static_cast<decltype(HvhClosure)*>(ci->hv1);
            (*HvhClosLoc)(x, n, dx, ndx, y, g, H);
            return 0;
          };
  }

  shandle oh = [](double const *x, int64_t n,
                  double const *y, int64_t m,
                  double const *, double const *,
                  char const **onames, char const **ovals, int64_t no,
                  void *arg) -> int {
        CallInfo *ci = static_cast<CallInfo*>(arg);
        auto ohClosLoc = static_cast<decltype(ohClosure)*>(ci->o1);
        int rc = (*ohClosLoc)(x, n, y, m, onames, ovals, no);
        return rc;
  };

  SEXP xs = PROTECT(allocVector(REALSXP, n));
  SEXP ys = PROTECT(allocVector(REALSXP, m));

  lastStatus0 = lastStatus1 = 0;

  int res = uopt(REAL(x0), REAL(xs), length(x0), REAL(ys), length(ys),
                 fh, gh, Jvh, Hh, Hvh, oh,
                 &myoptnames[0], &myoptvals[0], myoptvals.size(),
                 &mycallInfo);

  UNPROTECT(length(options) + 8);

  SEXP reslist     = PROTECT(allocVector(VECSXP, 3));
  SEXP resnames    = PROTECT(allocVector(VECSXP, 3));

  SET_VECTOR_ELT(reslist, 0, PROTECT(mkans(res)));

  SEXP sval = PROTECT(allocVector(STRSXP, 1));
  SET_STRING_ELT(sval, 0, mkChar("status"));
  SET_VECTOR_ELT(resnames, 0, sval);

  SET_VECTOR_ELT(reslist, 1, lastStatus0 ? lastStatus0 : mkans(0));

  sval = PROTECT(allocVector(STRSXP, 1));
  SET_STRING_ELT(sval, 0, mkChar("exit"));
  SET_VECTOR_ELT(resnames, 1, sval);

  SET_VECTOR_ELT(reslist, 2, lastStatus1 ? lastStatus1 : mkans(0));

  sval = PROTECT(allocVector(STRSXP, 1));
  SET_STRING_ELT(sval, 0, mkChar("last"));
  SET_VECTOR_ELT(resnames, 2, sval);

  setAttrib(reslist, R_NamesSymbol, resnames);

  setAttrib(xs, install("info"), reslist);

  UNPROTECT(6);

  return xs;
}

static const R_CallMethodDef R_CallDef[] = {
    {"uoptr", (DL_FUNC) &wrapuopt, 9},
    {NULL, NULL, 0}
};

extern "C" void
attribute_visible  // optional
R_init_uoptr(DllInfo *dll)
{
    R_registerRoutines(dll, NULL, R_CallDef, NULL, NULL);
    R_useDynamicSymbols(dll, FALSE);
    R_forceSymbols(dll, TRUE);
}

extern "C" void
attribute_visible  // optional
R_unload_uoptr(DllInfo *)
{
}
